/* (c) copyright 2013-2015 Richard D. Lindsley, Brigham Young University */
/*****************************************************************
  Filename:  l1b_info.c

  Reads ASCAT L1B file and displays some information on it

  v0.5 RDL 2013-03-28:
      Initial version
  v0.6 RDL 2013-03-28:
      Detects whether the file is gzipped or not (can't handle gzipped files yet)
  v0.7 RDL 2014-01-23:
      Optionally dump a specified MDR

  Roadmap:
      * Use zlib to transparently read in gzipped L1B files? (Use find_l2b_match.c to help)
      * When I dump an MDR, maybe I can use a better format, like csv or yaml or something. Maybe even an HDF dump option.

******************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <argp.h>
#include <sys/stat.h>

#include "ascat_l1b_reader.h"

// Globals
const char *argp_program_bug_address = "<rlindsley@mers.byu.edu>";
const char *argp_program_version = "l1b_info version 0.7";
unsigned int VERBOSE = 0;

// Structs
struct prog_opts {
    char *in_l1b;               // Filename to input L1B
    int mdr_dump;               // Which MDR to dump
    unsigned int mode;          // Mode to operate in. Currently unused
};

// Helpful macros
#define FAIL { printf("*** fail in %s, %s() at line %d\n", __FILE__, __FUNCTION__, __LINE__); exit(EXIT_FAILURE); }
#define WARN { printf("*** warning at line %i, continuing\n",__LINE__); }
#define FREE(p)   do { free(p); (p) = NULL; } while(0)

/********************************************************************/
// Function prototypes
void print_mdr(ascat_file_t *l1b_file, ascat_node_t *l1b_node);
static inline void clearline();
static void Dprintf(const unsigned int level, const char* format, ...);

/********************************************************************/
/*                Main helper functions                            */
/********************************************************************/

static int parse_opt(int key, char *arg, struct argp_state *state) {
    struct prog_opts *p_opts = state->input;

    switch(key) {
        case ARGP_KEY_INIT:
            p_opts->mode = 0;
            p_opts->in_l1b = NULL;
            p_opts->mdr_dump = -1;
            break;
        case 'v':
            VERBOSE = 1;
            break;
        case 'h':
            argp_state_help(state, stdout, ARGP_HELP_STD_HELP);
            break;
        case 'm':
            p_opts->mdr_dump = atoi(arg);
            if (p_opts->mdr_dump < 0)
                argp_failure(state, 1, 0, "Error in mdr_dump: %s cannot be negative", arg);
            break;
        case ARGP_KEY_ARG:
            if (p_opts->in_l1b == NULL) {
                p_opts->in_l1b = arg;
                // Check that the file exists (but not necessarily that it's a SIR file)
                if (access(arg, F_OK|R_OK) != 0)
                    argp_failure(state, 1, 0, "ERROR: %s doesn't exist or I don't have permission to read it", arg);
            }
            else {
                argp_state_help(state, stdout, ARGP_HELP_SHORT_USAGE);
                argp_failure(state, 1, 0, "Too many arguments");
            }
            break;
        case ARGP_KEY_END:
            if (p_opts->in_l1b == NULL) {
                argp_state_help(state, stdout, ARGP_HELP_SHORT_USAGE);
                argp_failure(state, 1, 0, "input-l1b not specified");
            }
            break;
    }

    return(0);
}

static int isGzipped(const char *fname) {
/* check if the file is gzipped or not
 * returns: 1 if gzipped
 *          0 if not
 *          -1 if error
 */
    FILE *imf;
    long int file_size;
    unsigned char magic_bytes[4];

    // Open file
    imf = fopen(fname, "r");
    if (imf == NULL) {
        fprintf(stderr,"*** ERROR: Can't open file %s\n", fname);
        exit(1);
    }

    // Read first two bytes to see if it's gzipped
    if (fread(&magic_bytes[0], sizeof(char), 1, imf) == 0) return(-1);
    if (fread(&magic_bytes[1], sizeof(char), 1, imf) == 0) return(-1);

    if ((magic_bytes[0] == 0x1f) && (magic_bytes[1] == 0x8b)) {
        // We're gzipped, so let's read the footer to find the uncompressed size
        fseek(imf, -4, SEEK_END);
        if (fread(&magic_bytes[0], sizeof(char), 1, imf) == 0) return(-1);
        if (fread(&magic_bytes[1], sizeof(char), 1, imf) == 0) return(-1);
        if (fread(&magic_bytes[2], sizeof(char), 1, imf) == 0) return(-1);
        if (fread(&magic_bytes[3], sizeof(char), 1, imf) == 0) return(-1);

        // Close file
        fclose(imf);

        file_size = magic_bytes[0] | magic_bytes[1] << 8 | magic_bytes[2] << 16 | magic_bytes[3] << 24;
        Dprintf(1, "Gzipped file detected\n");
        Dprintf(1, "Uncompressed file size: %ld bytes\n", file_size);
        return 1;
    }
    else {
        // Close file
        fclose(imf);
        return 0;
    }
}

/********************************************************************/
/*                Main function                                    */
/********************************************************************/
int main(int argc, char *argv[])
{
    struct prog_opts prog_opts;
    ascat_file_t l1b_file;
    ascat_orbit_t l1b_orb;
    ascat_node_t l1b_node;
    struct stat st;
    size_t tot_mdr;
    int imdr, mdr_err;

    //----
    // Parse arguments
    //----
    struct argp_option argp_options[] = {
        {0,0,0,0,"Output options", 9},
        {0, 'm', "MDR_NUM", 0, "Dump an MDR", 0},
        {0, 'v', 0, 0, "Verbose output", 0},
        {0,0,0,0,"Informational options", -1},
        {0, 'h', 0, OPTION_HIDDEN, "Help", 0},
        {0,0,0,0,0,0}
    };
    struct argp argp_in = {argp_options, parse_opt, "input-l1b", "Displays information on a given L1B file", 0, 0, 0};
    if (argp_parse(&argp_in, argc, argv, 0, 0, &prog_opts) != 0)
        return(1);

    Dprintf(1, "Input SIR: %s\n\n", prog_opts.in_l1b);

    //----
    // Read in L1B
    //----
    if (isGzipped(prog_opts.in_l1b)) {
        fprintf(stderr, "ERROR: file is gzipped, can't read it (yet)\n");
        return 1;
    }
    if (isatty(1)) Dprintf(1, "Checking ASCAT file...");
    if (ascat_check_file(prog_opts.in_l1b)) FAIL;
    if (isatty(1)) Dprintf(1, "done\n");

    if (isatty(1)) Dprintf(1, "Opening ASCAT file...");
    if (ascat_open_file(prog_opts.in_l1b, &l1b_file, &l1b_orb)) FAIL;
    if (isatty(1)) Dprintf(1, "done\n");

    //----
    // Display info
    //----
    fprintf(stdout, "L1B file: %s\n", prog_opts.in_l1b);
    fprintf(stdout, "  Product: %s\n", ascat_get_fmt(&l1b_file));
    fprintf(stdout, "  Satellite: ");
    switch (ascat_get_sat(&l1b_file)) {
        case METOP1:
            fprintf(stdout, "MetOp-B\n");
            break;
        case METOP2:
            fprintf(stdout, "MetOp-A\n");
            break;
        case METOP3:
            fprintf(stdout, "MetOp-C\n");
            break;
        default:
            fprintf(stdout, "Unknown\n");
            break;
    }

    fprintf(stdout, "\n");
    fprintf(stdout, "  L1B format version: %d\n", l1b_file.fver);
    stat(prog_opts.in_l1b, &st);
    fprintf(stdout, "  Size: %lu bytes (%lu MiB) \n", (unsigned long) st.st_size, (unsigned long) st.st_size / 1048576);

    fprintf(stdout, "  MDR size: %zu bytes each\n", l1b_file.mdr_size);
    tot_mdr = ascat_get_tot_mdr(&l1b_file);
    fprintf(stdout, "  MDRs: %zd\n", tot_mdr);
    fprintf(stdout, "  Nodes per beam: %zd\n", ascat_get_node_beam(&l1b_file));
    fprintf(stdout, "  Nodes in the file: %zd\n", ascat_get_tot_node(&l1b_file));

    //----
    // MDR dump
    //----
    if (prog_opts.mdr_dump >= 0) {
        fprintf(stdout, "Dumping MDR %d\n", prog_opts.mdr_dump);
        if ((size_t) prog_opts.mdr_dump > tot_mdr) {
            fprintf(stderr, "ERROR: Cannot dump MDR %d when only %zd MDRs exist in file\n", prog_opts.mdr_dump, tot_mdr);
            FAIL;
        }
        imdr = -1;
        do {
            mdr_err = ascat_next_mdr(&l1b_file, &l1b_node);
            if (mdr_err == 1) {
                fprintf(stderr, "ERROR in ascat_next_mdr, last MDR: %d\n", imdr);
                FAIL;
            } else if (mdr_err == 2) {
                printf("MDRs ran out\n");
            } else {
                imdr++;

                // We've reached the specified MDR
                if (imdr == prog_opts.mdr_dump) {
                    print_mdr(&l1b_file, &l1b_node);
                    break;
                }
            }
        } while(mdr_err == 0);

    }

    //----
    // Clean up
    //----
    ascat_close_file(&l1b_file);
    return(0);
}

/********************************************************************/
/*                Small helper functions                            */
/********************************************************************/
void print_mdr(ascat_file_t *l1b_file, ascat_node_t *l1b_node) {
    // Print the data from the current MDR
    int node_err = 0;
    int inode = 0;

    do {
        node_err = ascat_next_node(l1b_file, l1b_node);
        if (node_err == 1) {
            fprintf(stderr, "ERROR in ascat_next_node, last node: %d\n", inode);
            FAIL;
        } else if (node_err == 2) {
            //printf("nodes ran out (%d)\n", inode);
            ;
        } else {
            inode++;
            printf("node: %d swath: %d index: %d lat/lon: %f %f sigma0: %f inc: %f azi: %f\n", inode, l1b_node->swath, l1b_node->index, l1b_node->lat, l1b_node->lon, l1b_node->sigma0, l1b_node->inc, l1b_node->azi);
            //printf("node: %d swath: %d index: %d lat/lon: %f %f at prev lat/lon: %f %f at next lat/lon: %f %f\n", inode, l1b_node->swath, l1b_node->index, l1b_node->lat, l1b_node->lon, l1b_node->track_prev_lat, l1b_node->track_prev_lon, l1b_node->track_next_lat, l1b_node->track_next_lon);
        }
    } while (node_err == 0);

    return;
}

void Dprintf(const unsigned int level, const char* format, ...) {
  // Debug-printf
  // Only printf's if level >= VERBOSE
  // Adapted from <http://www.ozzu.com/cpp-tutorials/tutorial-writing-custom-printf-wrapper-function-t89166.html>
  va_list args;
  // 5 is debug level
  if (level <= VERBOSE) {
    if (level >= 5) {
      fprintf(stdout, "DEBUG: ");
    }
    va_start(args, format);
    vfprintf(stdout, format, args);
    va_end(args);
  }
  //fprintf(stdout, "\n");
  //NB, I could use stderr instead of stdout but this does weird things when I pipe to less
  return;
}

void clearline() {
  printf("\033[2K"); // ANSI escape sequence to clear the line right of cursor
  return;
}


// vim:softtabstop=4:tabstop=4:shiftwidth=4:
